// -------------------------------------------------------------------------------------------------
// =======
// Quantum
// =======
// Author: David414
// License: RWPCL (https://robowiki.net/wiki/RWPCL)
// -------------------------------------------------------------------------------------------------
package d414.nano;

import robocode.*;
import robocode.util.Utils;

import java.awt.Color;

public class Quantum extends AdvancedRobot {
    // ---------------------------------------------------------------------------------------------
    // Tuning Knobs
    // ---------------------------------------------------------------------------------------------
    static final double WALL_FORCE = 1;
    static final double AGRAV_DECAY = 0.9;
    static final double AHEAD_AMOUNT = 120;

    static final double AIM_START = 10;
    static final double AIM_FACTOR = 1.008;
    static final double RADAR_LOCK_THRESHOLD = 1;
    static final int    POWER_FACTOR = 2;

    // ---------------------------------------------------------------------------------------------
    // Globals
    // ---------------------------------------------------------------------------------------------
    static String targetName;
    static double targetDistance;
    static double xForce;
    static double yForce;

    @Override
    public void run() {
        setAllColors(Color.cyan);
        setAdjustGunForRobotTurn(true);
        onRobotDeath(null);
    }

    @Override
    public void onRobotDeath(RobotDeathEvent e) {
        setTurnRadarRight(targetDistance = Double.POSITIVE_INFINITY);
    }

    @Override
    public void onScannedRobot(ScannedRobotEvent e) {
        double distance;
        double aimAngle;
        double absoluteBearing;

        setTurnRightRadians(Utils.normalRelativeAngle(
                    Math.atan2(
                        (xForce = xForce * AGRAV_DECAY
                         - Math.sin(
                             aimAngle = (absoluteBearing = e.getBearingRadians() + getHeadingRadians())
                             + (e.getVelocity()
                                 / (AIM_START + Math.pow(AIM_FACTOR, distance = e.getDistance()))
                               )
                             * Math.sin(e.getHeadingRadians() - absoluteBearing)
                             ) / distance
                        )
                        + calcWallForce(getX())
                        , 
                        (yForce = yForce * AGRAV_DECAY - Math.cos(aimAngle) / distance)
                        + calcWallForce(getY())
                        )
                    - getHeadingRadians()
                    ));
        setAhead(AHEAD_AMOUNT - Math.abs(getTurnRemaining()));
        
        if (distance < targetDistance || e.getName() == targetName) {
            targetName = e.getName();
            targetDistance = distance;

            if (getGunHeat() < RADAR_LOCK_THRESHOLD) {
                setTurnRadarLeft(getRadarTurnRemaining());
            }

            setFire(getOthers() / POWER_FACTOR);

            setTurnGunRightRadians(Utils.normalRelativeAngle(aimAngle - getGunHeadingRadians()));
        }
    }

    public static double calcWallForce(double d) {
        return (WALL_FORCE / d) - (WALL_FORCE / (1000 - d));
    }
}
