/*
FoilistNano v3.1 by Sheldor.  03/23/2024  248 bytes
a nanobot with random movement and pattern matching targeting
v3.1 -- energy management, distancing

Foil is one of the three forms of modern sport fencing,
along with épée and sabre.  https://en.wikipedia.org/wiki/Foil_%28fencing%29

Credits:
Thanks go to the authors of the following bots: 
jk.sheldor.nano.Yatagan, jk.micro.Toorkild, jk.micro.Cotillion, and oog.nano.Fuatisha.
Also, a general thanks to all open source bot authors and contributors to the RoboWiki.

FoilistNano is open source and released under the terms of the RoboWiki Public Code License (RWPCL) - Version 1.1.
See license here: http://robowiki.net/wiki/RWPCL
*/

package sheldor.nano;

import robocode.*;
import robocode.util.Utils;
import java.lang.Math;

public class FoilistNano extends AdvancedRobot
{
	//global variables
	static double direction = Double.POSITIVE_INFINITY;
	
	public void onStatus(StatusEvent e)
	{		
		//set the radar and gun to turn independently
		//Putting these here with the radar code lets me get rid of the run() method, saving one byte.  credit to Cotillion
		setAdjustRadarForGunTurn(true);		
		setAdjustGunForRobotTurn(true);

		//turn the radar every tick
		//Putting the code here instead of in a while(true) loop in the run() method saves one byte.
		//I believe Wompi discovered this.
		setTurnRadarRightRadians(1);
	}	

	public void onScannedRobot(ScannedRobotEvent e) 
	{		
		//local variables
		int integer = 40;
       double absoluteBearing;
		int distance;
       int matchPosition;	
		
		//distancing based on Yatagan's
		setTurnRightRadians(Math.cos((absoluteBearing = e.getBearingRadians())
			+ (220 - (distance = (int)e.getDistance())) * (getVelocity() / 8000)));
		
		//pattern matching based on Yatagan's
		//Record the current enemy lateral velocity.
		enemyHistory = String.valueOf((char)Math.round(e.getVelocity() * (Math.sin(e.getHeadingRadians() - (absoluteBearing += getHeadingRadians()))))).concat(enemyHistory);
		  
		//radar
		setTurnRadarRightRadians(Utils.normalRelativeAngle(absoluteBearing - getRadarHeadingRadians()));

        //Find the longest match allowable in the enemy history.
        do{}while((matchPosition = enemyHistory.indexOf(enemyHistory.substring(0, --integer), 64)) < 0);			
		
		//energy management based on distance and enemy energy		
		setFire(Math.min((7.0 / 3.0) + (integer = (127 / distance)), e.getEnergy() / 4));
		
		//random movement inspired by Yatagan
		//Don't randomly reverse direction if the enemy is very close.
		setAhead(direction *= (Math.random() + integer - 0.0875));
		
		//Assume that the enemy will repeat their movements in the ticks following the recording of the original pattern.
		//Generate a firing angle based on that assumption.
		do
		{
			//Go through the log of enemy movements right after the original pattern was recorded.
			//Add the angular velocities directly to the absoluteBearing variable.
        	absoluteBearing += ((short) enemyHistory.charAt(--matchPosition)) /  e.getDistance();
        }		
		while ((distance -= 13) > 0); //Keep projecting the enemy's movement forward until our bullet would reach their predicted location.
		
		//Aim the gun at the enemy's predicted location.
		setTurnGunRightRadians(Utils.normalRelativeAngle(absoluteBearing - getGunHeadingRadians()));
	}	
	
	public void onHitWall(HitWallEvent e) 
	{
		//Reverse direction when the bot hits a wall.
		direction = -direction;
	}	
	
	//symbolic log of enemy movements for pattern matcher, from Yatagan
	//preloaded to prevent StringIndexOutOfBoundsException
   static String enemyHistory = ""
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 1 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 1 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 1 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 1 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 + (char) 0 
   + (char)-1 + (char)-2 + (char)-3 + (char)-4 + (char)-5 + (char)-6 
   + (char)-7 + (char)-8 + (char) 8 + (char) 7 + (char) 6 + (char) 5 
   + (char) 4 + (char) 3 + (char) 2 + (char) 1 + (char) 0 + (char) 0
   ;
}																																																												 																																										